#!/usr/bin/python

'''
PyWeek #8: 'Get off my lawn'

pyglet is by Alex Holkner and co.

sound made in lmms and edited in Audacity

open 2009 by george oliver
'''

###############################
# import
###############################

import random, math, os

import pyglet
from pyglet.window import key
from pyglet.gl import * 


###############################
# resources
###############################

pyglet.resource.path = ['data/sounds', 'data/images']
pyglet.resource.reindex()



###############################
# utility
###############################

# get more speed out of OpenGL, courtesy of cyhawk@PyWeek

pyglet.options['debug_gl'] = False
def zero_errcheck( result, func, arguments ):
	return result
pyglet.gl.lib.errcheck = zero_errcheck



def vector_add(vector1, vector2):
    v1_x, v1_y = vector1
    v2_x, v2_y = vector2
    
    v_x = v1_x + v2_x
    v_y = v1_y + v2_y
    
    return (v_x, v_y)

def vector_subtract(vector1, vector2):
    v1_x, v1_y = vector1
    v2_x, v2_y = vector2
    
    v_x = v1_x - v2_x
    v_y = v1_y - v2_y
    
    return (v_x, v_y)
    
def vector_multiply(vector1, scalar):
    v1_x, v1_y = vector1
    
    v_x = v1_x * scalar
    v_y = v1_y * scalar
    
    return (v_x, v_y)
    
def vector_divide(vector1, scalar):
    v1_x, v1_y = vector1
    
    v_x = v1_x//scalar
    v_y = v1_y//scalar
    
    return (v_x, v_y)
    
def vector_distance(vector1, vector2):
    v1_x, v1_y = vector1
    v2_x, v2_y = vector2
    
    x_sum = ((v2_x - v1_x) * (v2_x - v1_x))
    y_sum = ((v2_y - v1_y) * (v2_y - v1_y))
    sum = x_sum + y_sum
    distance = int(math.sqrt(sum))
    
    return distance
    
    
    
###############################
# window
###############################

window = pyglet.window.Window(800, 600, style='borderless', visible=False)

window.set_location(
                    window.screen.width//2-window.width//2,
                    window.screen.height//2-window.height//2)



###############################
# batches and groups
###############################

batch = pyglet.graphics.Batch()
hide = pyglet.graphics.Batch()

layer_0 = pyglet.graphics.OrderedGroup(0)
layer_1 = pyglet.graphics.OrderedGroup(1)
layer_2 = pyglet.graphics.OrderedGroup(2)

keys = key.KeyStateHandler()
window.push_handlers(keys)




###############################
# sound
###############################

class Sound(object):
    def __init__(self):
        for file in os.listdir('./data/sounds'):
            name = file.split('.')[0]
            self.__dict__[name] = pyglet.resource.media(file)
            
    def loop(self, sound):
        name = sound + '_loop'
        self.__dict__[name] = pyglet.media.Player()
        self.__dict__[name].queue(self.__dict__[sound])
        self.__dict__[name].eos_action = 'loop'
        self.__dict__[name].play()
        
sound = Sound()



###############################
# classes
###############################

class LetterImagePattern(pyglet.image.ImagePattern):
    def __init__(self, color=(150, 150, 150, 255)):
        self.color = '%c%c%c%c' % color

    def p(self):
        line_1 = self.color * 100
        line_2 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 80) + self.color * 10
        line_3 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 90)
        data = (line_3*40)+(line_1*10) + (line_2*40) + (line_1*10)
        return pyglet.image.ImageData(100, 100, 'RGBA', data)
        
    def e(self):
        line_1 = self.color * 100
        line_2 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 90)
        line_3 = (self.color * 90) + ('%c%c%c%c' % (0, 0, 0, 0) * 10)
        data = (line_1*10) + (line_2*30) + (line_3*10) + (line_2*40) + (line_1*10)
        return pyglet.image.ImageData(100, 100, 'RGBA', data)
        
    def n(self):
        line_1 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 80) + self.color * 10
        line_2 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 70) + self.color * 20
        line_3 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 60) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 10) + self.color * 10
        line_4 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 50) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 20) + self.color * 10
        line_5 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 40) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 30) + self.color * 10
        line_6 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 30) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 40) + self.color * 10
        line_7 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 20) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 50) + self.color * 10
        line_8 = self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 10) + self.color * 10 + ('%c%c%c%c' % (0, 0, 0, 0) * 60) + self.color * 10
        line_9 = self.color * 20 + ('%c%c%c%c' % (0, 0, 0, 0) * 70) + self.color * 10
        data = line_1*10 + line_2*10 + line_3*10 + line_4*10 + line_5*10 + line_6*10 + line_7*10 + line_8*10 + line_9*10 + line_1*10
        return pyglet.image.ImageData(100, 100, 'RGBA', data)
       
class Letter(pyglet.sprite.Sprite):
    def __init__(self, letter, x, y):
        letters = {
                    'p' : LetterImagePattern().p(),
                    'e' : LetterImagePattern().e(),
                    'n' : LetterImagePattern().n()}
        image = letters[letter]
        super(Letter, self).__init__(image, x, y, batch=batch, group=layer_2)
        


class HollowRectangleImagePattern(pyglet.image.ImagePattern):
    def __init__(self, color=(0, 0, 0, 0)):
        self.color = '%c%c%c%c' % color

    def create_image(self, width, height, line_width):
        top = self.color * width * line_width
        bottom = self.color * width * line_width
        middle = (self.color * line_width) + ('%c%c%c%c' % (0, 0, 0, 0) * (width - line_width*2)) + (self.color * line_width)
        data = top + (middle * (height - line_width*2)) + bottom
        return pyglet.image.ImageData(width, height, 'RGBA', data)

class Player(pyglet.sprite.Sprite):
    def __init__(self, x, y):
        hollow_rectangle = HollowRectangleImagePattern((150, 150, 150, 255))
        image = hollow_rectangle.create_image(100, 100, 10)
        image.anchor_x = image.width//2
        image.anchor_y = image.height//2
        super(Player, self).__init__(image, x, y, batch=batch, group=layer_2)
        self.speed = 500
        self.open = False
        
    def update(self, dt):
        if keys[key.UP]: self.y += self.speed * dt; self.collision_check(dt, 'up')
        if keys[key.DOWN]: self.y -= self.speed * dt; self.collision_check(dt, 'down')
        if keys[key.RIGHT]: self.x += self.speed * dt; self.collision_check(dt, 'right')
        if keys[key.LEFT]: self.x -= self.speed * dt; self.collision_check(dt, 'left')
        if keys[key.SPACE]: self.open = True; self.color=(0, 255, 255)
        if keys[key.X]: self.open = False; self.color=(255, 255, 255); self.paint()
        if keys[key.V]: print 'clear'
            
    def collision_check(self, dt, direction):
        for other in things.others:
            other_top = (other.y - (other.height//2)) + other.collision_height
            other_bottom = (other.y - (other.height//2)) + other.y_offset
            other_left = (other.x - (other.width//2)) + other.x_offset
            other_right = (other.x - (other.width//2)) + other.collision_width

            self_top = (self.y - (self.height//2)) + self.height
            self_bottom = self.y - self.height//2
            self_left = self.x - self.width//2
            self_right = (self.x - (self.width//2)) + self.width
            
            if self_top < other_bottom: pass
            if self_bottom > other_top: pass
            if self_left > other_right: pass
            if self_right < other_left: pass
            else: 
                if self.open: pass
                elif direction == 'up': other.y += self.speed * dt
                elif direction == 'down': other.y -= self.speed * dt
                elif direction == 'right': other.x += self.speed * dt
                elif direction == 'left': other.x -= self.speed * dt
                    
    def paint(self):
        for other in things.others:
            if (((self.x-self.width//2) < other.x < (self.x+self.width//2)) and
                ((self.y-self.height//2) < other.y < (self.y+self.height//2))):
                        x = int(other.x)
                        y = int(other.y)
                        vertex_list = batch.add(4, pyglet.gl.GL_QUADS, layer_2, 
                        ('v2i', (x, y,  x+8, y,  x+8, y+8,  x, y+8)), 
                        ('c3B', (255, 0, 0,  255, 0, 0,  255, 0, 0,  255, 0, 0)))
                    
               


class Block(pyglet.image.SolidColorImagePattern):
    def __init__(self, color):
        super(Block, self).__init__(color)
        self.image = self.create_image(8, 8)
        self.image.anchor_x = self.image.width//2
        self.image.anchor_y = self.image.height//2

class Other(pyglet.sprite.Sprite):
    def __init__(self, x, y, color=(255, 255, 255, 255), id='None'):
        block = Block(color)
        image = block.image
        self.id = id
        super(Other, self).__init__(image, x, y, batch=batch, group=layer_2)
        self.velocity = (0, 0)
        
        self.collision_width = self.width * 1
        self.collision_height = self.height * 1
        self.x_offset = (self.width - self.collision_width)//2
        self.y_offset = (self.width - self.collision_height)//2
        
    def update(self, dt):
        self.neighbors_find()
        
        v1 = self.cohere()
        v2 = self.separate()
        v3 = self.align()
        v4 = self.seek()
        v5 = self.rein()
        v6 = self.avoid()
                
        v = v1, v2, v3, v4, v5, v6
        for i in range(len(v)):
            self.velocity = vector_add(self.velocity, v[i])
        self.velocity = vector_multiply(self.velocity, dt*random.randint(4,12))
        self.position = vector_add(self.position, self.velocity)
        
        if not things.player.open: 
            if self.collision_check():
                self.position = vector_subtract(self.position, self.velocity)
                
        self.rotation += 10
            
    def neighbors_find(self):
        self.neighbors = []
        for other in things.others:
            if other.position != self.position:
                if ((self.x-32 < other.x < self.x+32) and
                    (self.y-32 < other.y < self.y+32)):
                        self.neighbors.append(other)
                
    def collision_check(self):
        self_top = (self.y - (self.height//2)) + self.collision_height
        self_bottom = (self.y - (self.height//2)) + self.y_offset
        self_left = (self.x - (self.width//2)) + self.x_offset
        self_right = (self.x - (self.width//2)) + self.collision_width
        
        player_top = (things.player.y - things.player.height//2) + things.player.height
        player_bottom = things.player.y - things.player.height//2
        player_left = things.player.x - (things.player.width//2)
        player_right = (things.player.x - (things.player.width//2)) + things.player.width

        if self_top < player_bottom: return False
        if self_bottom > player_top: return False
        if self_left > player_right: return False
        if self_right < player_left: return False
        else: return True
            
    def cohere(self):
        perceived_center = (0, 0)
        
        for other in things.others:
            if other.position != self.position:
                perceived_center = vector_add(perceived_center, other.position)
        perceived_center = vector_divide(perceived_center, len(things.others)-1)
        
        return vector_divide(vector_subtract(perceived_center, self.position), 100)
        
    def separate(self):
        dv = (0, 0)
        
        for other in self.neighbors:
            if other.position != self.position:
                if math.fabs(vector_distance(other.position, self.position)) < 24:
                    dv = vector_subtract(dv, vector_subtract(other.position, self.position))
                    
        return dv
        
    def align(self):
        perceived_velocity = (0, 0)
        
        for other in things.others:
            if other.position != self.position:
                perceived_velocity = vector_add(perceived_velocity, other.velocity)
        perceived_velocity = vector_divide(perceived_velocity, len(things.others)-1)
        
        return vector_divide(perceived_velocity, 8)
                
    def avoid(self):
        return vector_multiply(vector_subtract(things.player.position, self.position), -.08)
        
    def seek(self):        
        return vector_divide(vector_subtract(things.goal, self.position), 10)
    
    def rein(self):
        v_x, v_y = 0, 0
        
        if self.x < 0:
            v_x = 100
        elif self.x > 800:
            v_x = -100
        if self.y < 0:
            v_y = 100
        elif self.y > 600:
            v_y = -100
            
        return (v_x, v_y)



background = pyglet.sprite.Sprite(pyglet.resource.image('open_background.png'), batch=batch, group=layer_0) 

        
        

class Thing(object):
    def __init__(self):
        self.table = {
                    'player' : Player(-540, 350),
                    'p' : Letter('p', -460, 300),
                    'e' : Letter('e', -330, 300),
                    'n' : Letter('n', -200, 300),
                    'others' : [],
                    'tiles' : [],
                    'goal' : (400, 300)}
        
    def create(self, thing):
        self.__dict__[thing] = self.table[thing]
        return self.__dict__[thing]
        
        
        
###############################
# states
###############################

class IntroState(object):
    def __init__(self):
        self.letters = [
                    things.create('player'),
                    things.create('p'), 
                    things.create('e'),
                    things.create('n')]
                    
        self.instructions = pyglet.text.Label(
                                                text="[SPACE] [ARROWS] [X]",
                                                font_name='Courier',
                                                font_size=20,
                                                bold=True,
                                                color=(150, 150, 150, 255),
                                                x=-360,
                                                y=230,
                                                anchor_x='center',
                                                batch=batch,
                                                group=layer_2)

    def update(self, dt):
        if things.n.x < 600:
            for letter in self.letters:
                letter.x += 100 * dt
            self.instructions.x += 100 * dt
        elif things.n.x > 598:          
            things.p.x += 100 * dt
            things.e.x += 150 * dt
            things.n.x += 175 * dt
            self.instructions.x += 175 * dt
        if things.p.x > 810:
            self.instructions.delete()
            states.append(PlayState())


class PlayState(object):
    def __init__(self):
        things.create('others')        
        for i in range(10):
            x = random.randint(-800, -10)
            y = random.randint(-600, -10)
            things.others.append(Other(x, y))
        things.others.append(Other(-100, -100, (255, 0, 0, 255), 'red'))
            
        things.create('goal')
        
        sound.loop('open_beat')
            
    def update(self, dt):
        if random.randint(0, 100) > 95:
            x = random.randint(100, 700)
            y = random.randint(100, 500)
            things.goal = (x, y)
        for other in things.others:
            other.update(dt)
        things.player.update(dt)



###############################
# main loop and events
###############################

def update(dt):
    if len(states):
        states[-1].update(dt)
    
@window.event
def on_draw():
    window.clear()
    batch.draw()
    
if __name__ == '__main__':
    window.clear()
    window.set_visible()
    things = Thing()
    states = []
    states.append(IntroState())
    pyglet.clock.schedule_interval(update, 1/60.0)
    pyglet.app.run()


